from __future__ import division
"""compile recipe flowchart in dot. Because I can't install luagraph
"""
import csv
from os import path
FOLDER = path.dirname(path.realpath(__file__))
RECIPE_PATH = path.join(FOLDER, 'recipe.csv')
DEMAND_PATH = path.join(FOLDER, 'demand.csv')
DOT_PATH = path.expanduser('~/Downloads/temp.dot')
EPSILON = 0.0001

modifiers = {'ass': 0.75, 'che': 1.25, 'fur': 2.0, 'min': 0.5, 'non': 1.0}
item_dict = dict()


class Item(object):
    def __init__(self, name, factory_type, material, time_cost):
        self.name = name
        self.modifier = modifiers[factory_type]
        if material == 'fluid':
            self.fluid = True
        else:
            self.fluid = False
        self.time_cost = time_cost
        self.sources = dict()
        self.fixed = False
        self.output = dict()
        self.sum_output = 0
        self.demand = 0

    def fix(self):
        if self.fixed:
            return True
        if len(self.output) > 0 and min(self.output.values()) == -1:
            return False
        self.fixed = True
        self.sum_output = sum(self.output.values())
        self.sum_output += self.demand
        if self.demand > 0:
            print('demand for {0}: {1}. Demand met'.format(
                self.name, self.demand))
        #print('\n{0} fixed! Cost = {1}'.format(self.name, self.sum_output))
        self.unfixed = [source_name for source_name in self.sources]
        self.unchecked = list()
        return True

    def check_upstream(self):
        if not self.fixed:
            raise Exception("you shouldn't check_upstream of unfixed items")
        accum_success = 0
        while len(self.unfixed) > 0 or len(self.unchecked) > 0:
            success = 0
            for source in self.unfixed:
                item_dict[source].output[self.name] =\
                    self.sum_output * self.sources[source]
                if item_dict[source].fix():
                    self.unfixed.remove(source)
                    self.unchecked.append(source)
                    success += 1
            for source in self.unchecked:
                result = item_dict[source].check_upstream()
                if result == -1:
                    self.unchecked.remove(source)
                else:
                    success += result
            if success == 0:
                return accum_success
            accum_success += success
        return -1

    def cost(self):
        return self.sum_output * self.time_cost / self.modifier


def dot_print(items):
    str_list = list()
    str_list.append('digraph factorio {\n    concentrate = false;\n')
    raw_nodes = list()
    for item in item_dict.values():
        if item.cost() < EPSILON and item.name != 'petrol':
            continue
        str_list.append('    {0}[label = "{0} {1:.2f}"];\n'.format(
                        item.name, item.cost()))
        if len(item.sources) == 0:
            raw_nodes.append(item.name)
        for target, volume in item.output.items():
            if target == 'output' or volume < EPSILON:
                continue
            if item.fluid:
                weight = 1000.0
            else:
                weight = volume * 100
            str_list.append(
                '    {0} -> {1} [weight = {2}, label={3:.2f}];\n'.format(
                    item.name, target, weight, volume))
    str_list.append('    {rank = same ' + ' '.join(raw_nodes) + '};\n')
    demand_list = [item.name for item in item_dict.values() if item.demand > 0]
    str_list.append('    {rank = same ' + ' '.join(demand_list) +
                    '};\n')
    str_list.append('}')
    fp = open(DOT_PATH, 'w')
    fp.writelines(str_list)
    fp.close()


def read_recipe(file_path):
    fp = open(file_path, 'r')
    data = csv.reader(fp, delimiter=',')
    data = [filter(None, line) for line in data]
    for line in data:
        line_iter = iter(line)
        new_item = Item(next(line_iter), next(line_iter), next(line_iter),
                        float(next(line_iter)))
        for key in line_iter:
            new_item.sources[key] = float(next(line_iter))
        item_dict[new_item.name] = new_item
    raw_nodes = [item.name for item in item_dict.values()
                 if len(item.sources) == 0]
    return (item_dict, raw_nodes)


def read_demand(file_path):
    fp = open(file_path, 'r')
    data = csv.reader(fp, delimiter=',', skipinitialspace=True)
    data = filter(None, next(data))
    demand_dict = dict()
    data_iter = iter(data)
    for key in data_iter:
        demand_dict[key] = float(next(data_iter))
    return demand_dict


def main():
    """actual procedure
    """
    # read items and relations from a csv file
    global item_dict, demand_dict
    item_dict, raw_nodes = read_recipe(RECIPE_PATH)
    demand_dict = read_demand(DEMAND_PATH)
    for item in demand_dict:
        item_dict[item].demand = demand_dict[item]
    "find output pointers"
    for name, item in item_dict.items():
        for key, value in item.sources.items():
            item_dict[key].output[name] = -1
    "get the exit points"
    exit_nodes = [item.name for item in item_dict.values()
                  if len(item.output) == 0]
    for key in exit_nodes:
        item_dict[key].fix()
    for key in exit_nodes:
        item_dict[key].check_upstream()
    for key in exit_nodes:
        item_dict[key].check_upstream()
    dot_print(item_dict)

if __name__ == '__main__':
    main()
